@@ -103,6 +103,9 @@ type Config struct {
 	CacheSizeBytes int    // Cache size (in bytes). Default: 64k
 	CacheMinTTL    uint32 // Minimum TTL for DNS entries (in seconds).
 	CacheMaxTTL    uint32 // Maximum TTL for DNS entries (in seconds).
+	// CacheOptimistic defines if the optimistic cache mechanism should be
+	// used.
+	CacheOptimistic bool
 
 	// Handlers (for the case when dnsproxy is used as a library)
 	// --
